/*
 *   Capricorn.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   from Emu42
//
//	 WorkerThread as main emulation loop, timers, debugger stuff 
//
//   IO preocessors are synchronous with Capricorn (after each Capricorn opcode
//   an iop processor can evolve
//
//   Same thing for HP9121 and HP2225 but at a slower pace (every 8 instructions)
//

#include "StdAfx.h"
#include "HP8x.h"
#include "mops.h"
#include "io.h"
#include "debugger.h"

#if defined HP87
#define CRT_SCAN 10217						// scan time 1/60 sec
#define CRT_DISP 827
#endif
#if defined HP85			
#define CRT_SCAN 10080						// scan time 1/61 sec
#define CRT_DISP 2268
#endif

#define SAMPLE    4474						// speed adjust sample frequency (see dwCapricornCycles)

#define r2(i) (*((WORD *)(Chipset.R+i)))
#define w2(r, d) *((WORD *)(Chipset.R+r)) = d

#define PC (*((WORD *)(Chipset.R+4)))

static INT   nDbgOldState = DBG_OFF;		// old state of debugger for suspend/resume
static BOOL  bCpuSlow = FALSE;				// enable/disable real speed
static DWORD dwOldCyc;						// cpu cycles at last event
static DWORD dwSpeedRef;					// timer value at last event
static DWORD dwTickRef;						// sample timer ticks
static DWORD dwHpFreq;						// high performance counter for timers
static DWORD dwHpCountms;					// counter value for 1ms timer
static DWORD dwHpCountns;					// counter value for 50s timer
static DWORD dwHpms;						// for 1ms timer
static DWORD dwHpns;						// for 50s timer
static BOOL  bDbgSkipInt = FALSE;			// execute interrupt handler

BOOL    bInterrupt = FALSE;					// no demand for interrupt
UINT    nState = SM_INVALID;				// actual system state
UINT    nNextState = SM_RUN;				// next initial system state
BOOL    bRealSpeed = FALSE;					// real speed emulation activated ?
BOOL    bSound = FALSE;						// sound emulation activated ?
BOOL    bSetDate = FALSE;					// set Time & Date activated ?
BOOL    bKeySlow = FALSE;					// slow down for key emulation
CHIPSET Chipset;
DWORD   dwCapricornCycles = 137;			// Capricorn cpu cycles in interval for real speed (at 4474 Hz)
// variables for debugger engine
HANDLE  hEventDebug;						// event handle to stop cpu thread
BOOL    bDbgAutoStateCtrl = TRUE;			// debugger suspend control by SwitchToState()
INT     nDbgState = DBG_OFF;				// state of debugger
DWORD   dwDbgStopPC = -1;					// stop address for goto cursor
WORD    dwDbgRstkp;							// stack recursion level of step over end
WORD    dwDbgRstk;							// possible return address
DWORD   *pdwInstrArray = NULL;				// last instruction array
WORD    wInstrSize;							// size of last instruction array
WORD    wInstrWp;							// write pointer of instruction array
WORD    wInstrRp;							// read pointer of instruction array
BYTE    byLastOp;							// last opcode (skip ARP & DRP) on debug

DWORD   dwMaxTimers = 0;					// Number of call to DoTimers before something ticks
WORD    wCyclesTimer = 100;					// Number of capricorn cycles before calling DoTimers


//################
//#
//#    Low level subroutines
//#
//################

//
// save last instruction in circular instruction buffer
//
static __inline VOID SaveInstrAddr(DWORD dwAddr)
{
	if (pdwInstrArray)						// circular buffer allocated
	{
		pdwInstrArray[wInstrWp] = dwAddr;
		wInstrWp = (wInstrWp + 1) % wInstrSize;
		if (wInstrWp == wInstrRp)
			wInstrRp = (wInstrRp + 1) % wInstrSize;
	}
}

static __inline VOID Debugger(VOID)			// debugger part
{
	LARGE_INTEGER lDummyInt;				// sample timer ticks
	BOOL bStopEmulation;

	SaveInstrAddr(PC);						// save pc in last instruction buffer

	// check for code breakpoints
	bStopEmulation = CheckBreakpoint(PC, 1, BP_EXEC);

	// check for step cursor
	bStopEmulation |= (dwDbgStopPC == PC);

	// step over interrupt execution
	if (bDbgSkipInt && !bStopEmulation && !Chipset.inte)
		return;

	// check for step into
	bStopEmulation |= (nDbgState == DBG_STEPINTO);

	// check for step over
	bStopEmulation |= (nDbgState == DBG_STEPOVER) && (dwDbgRstkp == (Chipset.R[6]|(Chipset.R[7]<<8)));

	// check for step out, something was popped from return stack (not good)
	if (nDbgState == DBG_STEPOUT && dwDbgRstkp == (Chipset.R[6]|(Chipset.R[7]<<8)))
	{
		_ASSERT(bStopEmulation == FALSE);
		if ((bStopEmulation = (PC == dwDbgRstk)) == FALSE)
		{
			dwDbgRstkp = (Chipset.R[6]|Chipset.R[7]<<8) - 2;	// save stack data
			dwDbgRstk = Nread_nio2(dwDbgRstkp);
		}
	}

	if (bStopEmulation)						// stop condition
	{
		// redraw debugger window and stop
		NotifyDebugger(TRUE);
		WaitForSingleObject(hEventDebug,INFINITE);

		// init slow down part
		dwOldCyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF);
		QueryPerformanceCounter(&lDummyInt);
		dwSpeedRef = lDummyInt.LowPart;
	}
}

//
// Adjust all timers: with capricorn cycles for real speed
//   with High Performance counter otherwise.
//   dtime is the time elapsed in capricorn cycles
//
static __forceinline VOID __fastcall DoTimers(DWORD dtime)
{
DWORD ioptime = dtime;		// ref elapsed time in capricorn cycles
static DWORD dwNbCalls = 0;

	if ( !(bCpuSlow || bKeySlow) && (nDbgState == DBG_OFF))	// compute elapsed time with High Performance counter ?
	{
		LARGE_INTEGER lAct;

		dwNbCalls++;
		QueryPerformanceCounter(&lAct);
		if ((lAct.LowPart - dwHpCountms) <= dwHpms)			// one ms elapsed ?
			dtime = 0;
		else
		{
			dtime = 613*(lAct.LowPart - dwHpCountms);			
			dtime /= dwHpms;									// number of capricorn cycles (613062.5 Hz)
			dwHpCountms = lAct.LowPart;
		}
		if ((lAct.LowPart - dwHpCountns) <= dwHpns)			// 50 s elapsed ?
			ioptime = 0;
		else
		{
			ioptime = 31*(lAct.LowPart - dwHpCountms);
			ioptime /= dwHpns;									// number of capricorn cycles
			dwHpCountns = lAct.LowPart;
			if (dwNbCalls > dwMaxTimers)						// something ticks
				dwMaxTimers = dwNbCalls;
			if (dwNbCalls > 2)									// number of calls 
				if (wCyclesTimer < 0xFFFF)
					wCyclesTimer++;									// too much, raise  
			else
				if (wCyclesTimer > 30)								
					wCyclesTimer--;									// not enough, lower
			dwNbCalls = 0;
		}
	}
	
	if (dtime > 0)
	{
		int i;

		Chipset.crttimer += (DWORD)(dtime);

#if defined HP85
		if (Chipset.crtcount > 0)
		{
			if (Chipset.crttimer >= CRT_SCAN)		// for service rom
			{
				Chipset.crtcount--;
			}
		}
		else
		{
			Chipset.crtrsts &= 0x7F;			// crt not busy
		}
#endif

		while (Chipset.crttimer >= CRT_SCAN)	// one CRT retrace (1/60 sec)
		{
			Chipset.crttimer -= CRT_SCAN;
			if (bSoundStop)						// should we stop the sound ?
				HaltWaveDo();						// do the sound fade out
		}

		for (i = 0; i < 4; i++)				// clk timers
		{
			if (Chipset.inttimer[i] && Chipset.etimer[i])			// timer want an interrupt ?
			{
				if (Chipset.inte)										// interrupt enable ?
					bInterrupt = TRUE;										// take the interrupt
			}
			if (Chipset.etimer[i] & Chipset.ctimer[i])				// counting timer
			{
				Chipset.dtimer[i] += (WORD)(dtime);
				while (Chipset.dtimer[i] >= 613)						// one ms elapsed ?
				{
					Chipset.dtimer[i] -= 613;								// minus One ms
					Chipset.timer[i]++;										// increment timer
					if (Chipset.timer[i] == Chipset.ptimer[i])				// equal preset ?
					{
						Chipset.inttimer[i] = TRUE;								// want an interrupt
						Chipset.timer[i] = 0;									// reload preset
					}
				}
			}
		}
	}
	if (ioptime > 0)
	{
		int i;

		for (i = 0; i < 11; i++)			// iop timers (just for some reset delay), not very efficient
		{
			if (Chipset.iop[i].iop)
			{
				while (Chipset.iop[i].ioptimer > 0)
				{
					if (Chipset.iop[i].ioptimer > ioptime) Chipset.iop[i].ioptimer -= ioptime;
						else Chipset.iop[i].ioptimer = 0;
				}
			}
		}
	}
}

//
// adjust emulation speed
//
static __forceinline VOID AdjustSpeed(VOID)		
{
	DWORD dwCycles,dwTicks;

	// cycles elapsed for next check
	if ((dwCycles = (DWORD) (Chipset.cycles & 0xFFFFFFFF) - dwOldCyc) >= (DWORD) T2CYCLES)
	{
		LARGE_INTEGER lAct;
		do
		{
			QueryPerformanceCounter(&lAct);
		}
		while((dwTicks = lAct.LowPart - dwSpeedRef) <= dwTickRef);

		// workaround for QueryPerformanceCounter() in Win2k,
		// if last command sequence took over 50ms -> synchronize
		if(dwTicks > 223 * dwTickRef)	// time for last commands > 50ms (223 at 4474 Hz)
		{											// new synchronizing
			dwOldCyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF);
			QueryPerformanceCounter(&lAct);			// get timer ticks
			dwSpeedRef = lAct.LowPart;				// save reference time
		}
		else
		{
			dwOldCyc += T2CYCLES;					// adjust cycles reference
			dwSpeedRef += dwTickRef;				// adjust reference time
		}
	}
}

//################
//#
//#    Public functions
//#
//################

VOID SuspendDebugger(VOID)
{
	// auto control enabled, emulation halted by debugger
	if (bDbgAutoStateCtrl && nDbgState > DBG_OFF)
	{
		nDbgOldState = nDbgState;			// save old state
		nDbgState = DBG_SUSPEND;			// suspend state
		SetEvent(hEventDebug);				// exit debugger
	}
}

VOID ResumeDebugger(VOID)
{
	// auto control enabled, debugger is suspended
	if (bDbgAutoStateCtrl && nDbgState == DBG_SUSPEND)
		nDbgState = nDbgOldState;			// set to old debugger state
}

//
// wait for idle basic state: main EXEC loop (ready for input)
//
BOOL WaitForIdleState(VOID)
{
	DWORD dwRefTime;

	SuspendDebugger();									// suspend debugger

	dwRefTime = timeGetTime();
	Chipset.idle = FALSE;
	Sleep(0);
	// wait for the Iddle state with 5 sec timeout
	while (((timeGetTime() - dwRefTime) < 5000L) && (!Chipset.idle))
		Sleep(0);

	if (Chipset.idle)								// idle ?
		return FALSE;									// idle state
	else
		ResumeDebugger();								// timeout, resume to debugger

	return TRUE;									// emulator was busy
}

//
// Set HP time and date
//
VOID SetHPTime(VOID)
{
	SYSTEMTIME ts, tt;
	DWORD      dw, dy, dd;

#if defined HP87
	BYTE	   *lpAdr = Chipset.Ram + 0x025B;	// date is a 3 bytes BCD integer at 825B (date$ in AP use yyddd format)
#endif
#if defined HP85
	BYTE	   *lpAdr = Chipset.Ram + 0x01B0;	// date is a 3 bytes BCD integer at 825B (date$ in AP use yyddd format)
#endif

	GetLocalTime(&ts);							// local time, _ftime() cause memory/resource leaks

	dd = (ts.wYear % 100L) * 1000L;				// year part of yyddd format

	tt.wYear = ts.wYear - 1;					// 31/12 of previous year
	tt.wDay = 31;
	tt.wMonth = 12;

	// calculate days: today from 01.01.1970
	dw = (DWORD) ts.wMonth;
	if (dw > 2)
		dw -= 3L;
	else
	{
		dw += 9L;
		--ts.wYear;
	}
	dw = (DWORD) ts.wDay + (153L * dw + 2L) / 5L;
	dw += (146097L * (((DWORD) ts.wYear) / 100L)) / 4L;
	dw +=   (1461L * (((DWORD) ts.wYear) % 100L)) / 4L;
	dw -= 719469L;

	// calculate days: 31/12 previous year from 01.01.1970
	dy = (DWORD) tt.wMonth;
	if (dy > 2)
		dy -= 3L;
	else
	{
		dy += 9L;
		--tt.wYear;
	}
	dy = (DWORD) tt.wDay + (153L * dy + 2L) / 5L;
	dy += (146097L * (((DWORD) tt.wYear) / 100L)) / 4L;
	dy +=   (1461L * (((DWORD) tt.wYear) % 100L)) / 4L;
	dy -= 719469L;

	dd += dw - dy;	// days from the begining of the year

	// set the date

	lpAdr[0] =  (BYTE)(dd % 10);			dd /= 10;
	lpAdr[0] |= (BYTE)((dd % 10) << 4);		dd /= 10;
	lpAdr[1] =  (BYTE)(dd % 10);			dd /= 10;
	lpAdr[1] |= (BYTE)((dd % 10) << 4);		dd /= 10;
	lpAdr[2] =  (BYTE)(dd % 10);			dd /= 10;
	lpAdr[2] |= (BYTE)((dd % 10) << 4);

	// calculate time in milliseconds
	dw = (DWORD) ts.wHour;
	dw = dw * 60L + (DWORD) ts.wMinute;
	dw = dw * 60L + (DWORD) ts.wSecond;
	dw = dw * 1000L + (DWORD) ts.wMilliseconds;

	// set the time in timer 0 
	Chipset.timer[0] = dw;
}

VOID SetSpeed(BOOL bAdjust)					// set emulation speed
{
	if (bAdjust)							// switch to real speed
	{
		LARGE_INTEGER lTime;				// sample timer ticks
		// save reference cycles
		dwOldCyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF);
		QueryPerformanceCounter(&lTime);	// get timer ticks
		dwSpeedRef = lTime.LowPart;			// save reference time
	}
	bCpuSlow = bAdjust;						// save emulation speed
}

UINT SwitchToState(UINT nNewState)
{
	UINT nOldState = nState;

	if (nState == nNewState) return nOldState;
	switch (nState)
	{
	case SM_RUN: // Run
		switch (nNewState)
		{
		case SM_INVALID: // -> Invalid
			nNextState = SM_INVALID;
			bInterrupt = TRUE;				// leave inner emulation loop
			SuspendDebugger();				// suspend debugger
			while (nState!=nNextState) Sleep(0);
			UpdateWindowStatus();
			break;
		case SM_RETURN: // -> Return
			DisableDebugger();				// disable debugger
			nNextState = SM_INVALID;
			bInterrupt = TRUE;				// leave inner emulation loop
			while (nState!=nNextState) Sleep(0);
			nNextState = SM_RETURN;
			SetEvent(hEventShutdn);
			WaitForSingleObject(hThread,INFINITE);
			UpdateWindowStatus();
			break;
		}
		break;
	case SM_INVALID: // Invalid
		switch (nNewState)
		{
		case SM_RUN: // -> Run
			nNextState = SM_RUN;
			// don't enter opcode loop on interrupt request
			bInterrupt = FALSE;
			ResumeDebugger();
			SetEvent(hEventShutdn);
			while (nState!=nNextState) Sleep(0);
			UpdateWindowStatus();
			break;
		case SM_RETURN: // -> Return
			DisableDebugger();				// disable debugger
			nNextState = SM_RETURN;
			SetEvent(hEventShutdn);
			WaitForSingleObject(hThread,INFINITE);
			break;
		}
		break;
	}
	return nOldState;
}

UINT WorkerThread(LPVOID pParam)
{
	LARGE_INTEGER lDummyInt;				// sample timer ticks
	BOOL bIntFound;							// interrupt found flag

	QueryPerformanceFrequency(&lDummyInt);	// init timer ticks
/*	{
		TCHAR buffer[256];
		int i;

		i = wsprintf(buffer,_T("Performance frequency : (%08lX, %08lX)\n"), lDummyInt.HighPart, lDummyInt.LowPart);
		buffer[i++] = 0;
		OutputDebugString(buffer);
	} */
	QueryPerformanceFrequency(&lDummyInt);
	lDummyInt.QuadPart /= 1000;				// calculate 1ms ticks
	dwHpms = lDummyInt.LowPart;
	QueryPerformanceFrequency(&lDummyInt);
	lDummyInt.QuadPart /= 20000;			// calculate 50s ticks
	dwHpns = lDummyInt.LowPart;
	QueryPerformanceFrequency(&lDummyInt);
	lDummyInt.QuadPart /= SAMPLE;			// calculate sample ticks
	dwTickRef = lDummyInt.LowPart;			// sample timer ticks
	_ASSERT(dwTickRef);						// tick resolution error

loop:
	while (nNextState == SM_INVALID)		// go into invalid state
	{
		nState = SM_INVALID;				// in invalid state
		WaitForSingleObject(hEventShutdn,INFINITE);
		if (nNextState == SM_RETURN)		// go into return state
		{
			nState = SM_RETURN;				// in return state
			return 0;						// kill thread
		}
	}
	while (nNextState == SM_RUN)
	{
		if (nState != SM_RUN)
		{
			nState = SM_RUN;
			UpdateMainDisplay(TRUE);
			UpdateAnnunciators();
			if (bSetDate)
				SetHPTime();					// update HP time & date
			// init speed reference
			dwOldCyc = (DWORD) (Chipset.cycles & 0xFFFFFFFF);
			QueryPerformanceCounter(&lDummyInt);
			dwSpeedRef = lDummyInt.LowPart;
			dwHpCountms = lDummyInt.LowPart;
			dwHpCountns = lDummyInt.LowPart;
		}
		byLastOp = 0x80;
		while (!bInterrupt)
		{
			WORD pc = PC;

			if ((nDbgState > DBG_OFF) && (byLastOp > 0x7F))		// debugger active and not ARP & DRP opcode
			{
				Debugger();

				// if suspended skip next opcode execution
				if (nDbgState == DBG_SUSPEND)
					continue;
			}
			
			// main EXEC loop in idle state ? (or idle in input)
			if (pc == 0x003A)
			{
				if (r2(14) == 0x00 || r2(14) == 0x04)
				{
					if (Chipset.idle)				// if already in idle state ?
						Sleep(1);						// free host CPU
					Chipset.idle = TRUE;
				}
			}
#if defined HP87
			// p-SYSTEM waiting for a key ?
			if (pc == 0xD812)
				Sleep(1);
#endif
 			pc = EvalOpcode(pc);			// execute opcode
			Chipset.cycles_reserved++;		// count instructions
			if ( !(bCpuSlow || bKeySlow) && (nDbgState == DBG_OFF))	// compute elapsed time with High Performance counter ?
			{
				if (Chipset.dcycles > wCyclesTimer)						// enough Capricorn cycles
				{															// do it !
					Chipset.cycles += (DWORD)(Chipset.dcycles);
					DoTimers(Chipset.dcycles);		
					Chipset.dcycles = 0;
				}
			}
			else													// slow mode
				if (Chipset.dcycles > 31)								// at least 31 Capricorn cycles
				{															// do it
					Chipset.cycles += (DWORD)(Chipset.dcycles);				
					DoTimers(Chipset.dcycles);
					Chipset.dcycles = 0;
				}
			if (Chipset.iop[0].iop)		IopCmd(&Chipset.iop[0]);
			if (Chipset.iop[1].iop)		IopCmd(&Chipset.iop[1]);
			if (Chipset.iop[2].iop)		IopCmd(&Chipset.iop[2]);
			if (Chipset.iop[3].iop)		IopCmd(&Chipset.iop[3]);
			if (Chipset.iop[4].iop)		IopCmd(&Chipset.iop[4]);
			if (Chipset.iop[5].iop)		IopCmd(&Chipset.iop[5]);
			if (Chipset.iop[6].iop)		IopCmd(&Chipset.iop[6]);
			if (Chipset.iop[7].iop)		IopCmd(&Chipset.iop[7]);
			if (Chipset.iop[8].iop)		IopCmd(&Chipset.iop[8]);
			if (Chipset.iop[9].iop)		IopCmd(&Chipset.iop[9]);
			if (Chipset.iop[10].iop)	IopCmd(&Chipset.iop[10]);
			if (Chipset.iop[7].iop == 1)
			{
				if ((Chipset.cycles_reserved & 0x07) == 0x00)
					DoHP9121();
				if ((Chipset.cycles_reserved & 0x0F) == 0x04)
					DoHP2225();
#if defined HP7470
				if ((Chipset.cycles_reserved & 0x0F) == 0x0C)
					DoHP7470();
#endif
			}
			if (Chipset.iop[10].iop == 2)
				DoHP82939();
			if (bCpuSlow || bKeySlow)			// emulation slow down
				AdjustSpeed();						// adjust emulation speed
		}
		bInterrupt = FALSE;					// be sure to reenter opcode loop
		bIntFound = FALSE;					// for 

		if (Chipset.intkbd)					// keyboard interrupt ?
		{
			WORD adr;

			_ASSERT(Chipset.inte);
			bIntFound = TRUE;
			Chipset.intkbd = FALSE;
			Chipset.dcycles += 15; 
			adr = r2(6);
			Nwrite(Chipset.R+4, adr, 2);
			adr += 2;
			w2(6, adr);
			adr = Nread_nio2(0x0004);
			w2(4, adr);
		}
		if (!bIntFound)
		{
			int i;

			for (i= 0; i < 4; i++)				// timer interrupt ?
			{
				if (Chipset.inttimer[i]) 
				{
					WORD adr;

					_ASSERT(Chipset.inte);
					bIntFound = TRUE;
					Chipset.inttimer[i] = FALSE;
					Chipset.dcycles += 15; 
					adr = r2(6);
					Nwrite(Chipset.R+4, adr, 2);
					adr += 2;
					w2(6, adr);
					adr = Nread_nio2((WORD)(0x0008 + (i<<1)));
					w2(4, adr);
					break;											// found the interrupt, leave loop
				}
			}
		}
		if (!bIntFound)
		{
			int i;

			for (i= 0; i < 11; i++)				// iop interrupt ?
			{
				if (Chipset.iop[i].intiop) 
				{
					WORD adr;

					_ASSERT(Chipset.inte);
					bIntFound = TRUE;
					Chipset.iop[i].intiop = FALSE;					// got interrupt
					Chipset.intrsc = (BYTE)(0x4A+i*2);				// write IOP source
					Chipset.dcycles += 15;
					adr = r2(6);									// get stack address
					Nwrite(Chipset.R+4, adr, 2);					// save pc on stack
					adr += 2;										// increment stack
					w2(6, adr);										// write back stack
					adr = Nread_nio2(0x0010);						// get vector address
					w2(4, adr);										// write it in pc
					break;											// found the interrupt, leave loop
				}
			}
		}
	}
	_ASSERT(nNextState != SM_RUN);

	goto loop;
	UNREFERENCED_PARAMETER(pParam);
}
